#ifndef lint
static char sccsid[] = "@(#)hist.c	1.2 2/19/89";
#endif

/*
 * Copyright (c) 1988 by Sam Leffler.
 * All rights reserved.
 *
 * This file is provided for unrestricted use provided that this
 * legend is included on all tape media and as a part of the
 * software program in whole or part.  Users may copy, modify or
 * distribute this file at will.
 */

#include <stdio.h>
#include <ctype.h>
#include "tiffio.h"

/*
 * Histogram info, one per
 * possible tile flags value.
 */
struct picHistogram {
	int	h_count;		/* raw appearance count */
	int	h_maxcount;		/* max count */
	int	h_maxlen;		/* max run length */
	int	h_totlen;		/* average run length */
	int	h_totcount;		/* average count */
} hstats[16];

#define	PT_FILL		0		/* fill input buffer */
#define PT_FULLDUMP	1		/* full dump, use count */
#define PT_FULLRUN	2		/* full run, length precedes pixel */
#define PT_PARTDUMP	3		/* part dump, alpha constant */
#define PT_PARTRUN	4		/* part run, alpha constant */

#define	howmany(x, y)	(((x)+((y)-1))/(y))

main(argc, argv)
	char *argv[];
{
	register TIFF *tif;
	TIFFDirectory *td;
	u_char *scanbuf, *pp;
	int bitspersample, scanline, samplesperbyte;
	register int cc, bytesperline, stride;
	int PostScript = 0, w, h, i, dirnum = -9999;
	unsigned width = -1, height = -1;
	int PicioHistogram();

	for (argc--, argv++; argc > 0 && argv[0][0] == '-'; argc--, argv++) {
		if (strcmp(argv[0], "-p") == 0) {
			PostScript++;
			continue;
		}
		if (isdigit(argv[0][1])) {
			dirnum = atoi(&argv[0][1]);
			continue;
		}
		if (strcmp(argv[0], "-w") == 0) {
			argc--, argv++;
			width = atoi(argv[0]);
			continue;
		}
		if (strcmp(argv[0], "-h") == 0) {
			argc--, argv++;
			height = atoi(argv[0]);
			continue;
		}
	}
	if (argc <= 0) {
		fprintf(stderr,
		   "usage: a.out [-p] [-#] [-w width] [-h height] TIFF-file\n");
		exit(-1);
	}
	tif = TIFFOpen(argv[0], "r");
	if (tif == NULL)
		exit(-1);
	do {
		if (dirnum != -9999) {
			dirnum--;
			if (dirnum > 0)
				continue;
			if (dirnum < 0)
				break;
		}
		if (!PostScript)
			TIFFPrintDirectory(tif, stdout, 1, 0, 0);
		td = tif->tif_dir;
		if (td->td_compression == COMPRESSION_PICIO) {
			bzero(hstats, sizeof (hstats));
			tif->tif_decoderow = PicioHistogram;
		}
		bitspersample = td->td_bitspersample;
		if (td->td_planarconfig == PLANARCONFIG_CONTIG)
			bitspersample *= td->td_samplesperpixel;
		scanline = howmany(bitspersample * td->td_imagewidth, 8);
		scanbuf = (u_char *)malloc(scanline);
		if (scanbuf == NULL) {
			printf("No space for scanline buffer\n");
			continue;
		}
		w = (width > td->td_imagewidth ? td->td_imagewidth : width);
		h = (height > td->td_imagelength ? td->td_imagelength : height);
		if (PostScript)
			printf("%d %d %d [%d 0 0 -%d 0 %d] {<\n",
			     w, h, td->td_bitspersample, w, h, h);
		samplesperbyte =
		    bitspersample == 1 ? 8 : bitspersample == 4 ? 2 : 1;
		w /= samplesperbyte;
		stride = (td->td_planarconfig == PLANARCONFIG_CONTIG &&
		    td->td_samplesperpixel > 1 ? td->td_samplesperpixel-1 : 0);
		while (h-- > 0 && TIFFReadScanline(tif, scanbuf) > 0) {
			if (!PostScript)
				continue;
			pp = scanbuf;
			for (cc = 0; cc < scanline && cc < w; cc++) {
				printf(" %02x",
				  td->td_photometric == PHOTOMETRIC_MINISWHITE ?
				    ~(*pp++) & 0xff : *pp++);
				if (((cc+1) % 24) == 0)
					putchar('\n');
				pp += stride;
			}
			putchar('\n');
		}
		if (PostScript)
			printf("\n>} image\n");
		free((char *)scanbuf);
		if (td->td_compression == COMPRESSION_PICIO)
			PrintStats();
	} while (TIFFReadDirectory(tif));
	if (tif)
		TIFFClose(tif);
}

char *picNames[16] = {
    "FILL",	"FULLDUMP",	"FULLRUN",	"PARTDUMP",
    "PARTRUN",	"#5",		"#6",		"#7",
    "#8",	"#9",		"#10",		"#11",
    "#12",	"#13",		"#14",		"#15",
};

PicioHistogram(tif)
	TIFF *tif;
{
	static char module[] = "PicioHistogram";
	register u_char *bp, *op, *sp, *ep;
	register int npixels, inc;
	TIFFDirectory *td = tif->tif_dir;
	short nc, flag, word, count, length;
	u_char spare[4];
	register struct picHistogram *hp;

	bp = (u_char *)tif->tif_rawcp;
	ep = bp + tif->tif_rawcc;
	op = (u_char *)tif->tif_scanline;
	npixels = td->td_imagewidth;
	nc = td->td_samplesperpixel;
	inc = tif->tif_bytespersample;
	while (bp < ep && npixels > 0) {
		word = (bp[1] << 8) | bp[0], bp += 2;
		count = (word & 0xfff) + 1;
		flag = word >> 12;
		hp = &hstats[flag];
		hp->h_count++;
		if (count > hp->h_maxcount)
			hp->h_maxcount = count;
		hp->h_totcount += count;
		switch (flag) {
		case PT_FULLRUN:
			while (count-- > 0) {
				/*
				 * The first byte holds a count. 
				 * The next bytes hold data to be
				 * repeated.
				 */
				if (inc == 2) {
					length = *bp++;
					length |= *bp++ << 8;
					length++;
				} else
					length = (*bp++) + 1;
				if (length > hp->h_maxlen)
					hp->h_maxlen = length;
				hp->h_totlen += length;
				bp += nc * inc;
				npixels -= length;
			}
			break;
		case PT_FULLDUMP:
			npixels -= count;
			bp += count * nc * inc;
			break;
		case PT_PARTRUN:
			bp += inc;
			while (count-- > 0) {
				/*
				 * The first byte is a count. 
				 * The next bytes are data to
				 * be repeated.
				 */
				if (inc == 2) {
					length = *bp++;
					length |= *bp++ << 8;
					length++;
				} else
					length = (*bp++) + 1;
				if (length > hp->h_maxlen)
					hp->h_maxlen = length;
				hp->h_totlen += length;
				bp += (nc-1) * inc;
				npixels -= length;
			}
			break;
		case PT_PARTDUMP:
			/*
			 * The fixed value (normally alpha) is
			 * stored in the first byte.
			 */
			bp += inc;
			npixels -= count;
			bp += count * (nc-1) * inc;
			break;
		default:
			TIFFError(module, "%s: Unknown flag 0x%x",
			    tif->tif_name, flag);
			return (0);
		}
	}
	tif->tif_rawcc -= bp - (u_char *)tif->tif_rawcp;
	tif->tif_rawcp = (char *)bp;
	if (npixels > 0) {
		TIFFError(module, "%s: Not enough data for scanline %d",
		    tif->tif_name, tif->tif_row);
		return (0);
	}
	return (1);
}

PrintStats()
{
	int total;
	register struct picHistogram *hp;

	total = 0;
	for (hp = hstats; hp < &hstats[16]; hp++)
		total += hp->h_count;
	printf("%-9s          Count        Length\n", "");
	printf("%-9s   Total  Max   Avg   Max   Avg\n", "Type");
	/* we ignore PT_FILL statistics */
	for (hp = &hstats[1]; hp < &hstats[16]; hp++) {
		if (hp->h_count == 0)
			continue;
		printf("%-9s: %6d %4d %5.1f  %4d %5.1f  %4.1f%%\n",
		    picNames[hp - hstats], hp->h_count,
		    hp->h_maxcount,
#define	nz(x)	((x) == 0 ? 1 : (x))
		    (float)hp->h_totcount / nz(hp->h_count),
		    hp->h_maxlen,
		    (float)hp->h_totlen / nz(hp->h_count),
		    (100.0 * hp->h_count) / total);
	}
}
